`timescale 1ns / 1ps
////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer:
//
// Create Date:    13:54:22 08/15/05
// Design Name:    
// Module Name:    packet_manager
// Project Name:   
// Target Device:  
// Tool versions:  
// Description:
//
// Dependencies:
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
////////////////////////////////////////////////////////////////////////////////
module packet_manager(clk, reset, OwnMAC, DestMAC, 
   startOrbit, finishOrbit, hlxId,  
	packet_send, packet_len, packet_sel,
	packet_real_sel, packet_real_read, packet_real_write, packet_real_addr, packet_real_we,packet_real_hd,
	packet_real_send, packet_real_len, 
	//packet_send_poll, packet_send_busy, packet_send_macbusy, mode_wait_send, 
	busy,
	histogram_data,histogram_addr,histogram_read_strobe,histogram_read_ack
	);																	 
    input clk, reset;
    input [47:0] OwnMAC;
    input [47:0] DestMAC;
	 input [31:0] startOrbit, finishOrbit;
	 input [7:0] hlxId;
    input packet_send;
    input [15:0] packet_real_read;
	 input [10:0] packet_len;
	 input [5:0] packet_sel;
    output [15:0] packet_real_write;
	 output [8:0] packet_real_addr;
	 output [5:0] packet_real_sel;
	 output packet_real_we, packet_real_send, packet_real_hd;
	 output reg [10:0] packet_real_len;
	 //input mode_wait_send;
	 output reg busy;
	 input [15:0] histogram_data;
	 output reg [15:0] histogram_addr;
	 input histogram_read_ack;
	 output histogram_read_strobe;
//	 output packet_send_poll;
	 //input packet_send_busy, packet_send_macbusy;

	reg [2:0] state = 3'h0;
	reg [15:0] header_data;
	wire [15:0] histogram_data_swapped;
	reg [8:0] header_addr;
	wire [8:0] copy_addr;
	reg [5:0] header_sel,packet_fifo = 6'h0;
	reg header_we;
	wire copy_write_strobe;
	wire header_done, copy_done;

parameter STATE_IDLE=3'h0;
parameter STATE_COPY=3'h1;
parameter STATE_HEADER=3'h2;
parameter STATE_XMIT=3'h3;
parameter STATE_WAIT=3'h4;
parameter PacketType=16'h6006;
parameter MAX_PACKET_COUNT=6'h7;

always @(posedge clk)
	busy<=(state!=STATE_IDLE);

assign histogram_data_swapped={histogram_data[7:0],histogram_data[15:8]};

assign packet_real_write=(state==STATE_COPY)?(histogram_data_swapped):(header_data);
assign packet_real_addr=(state==STATE_COPY)?(copy_addr):(header_addr); // this will change when we make CRCs
assign packet_real_sel=packet_fifo;
assign packet_real_hd=(state==STATE_HEADER)?(1'h1):(1'h0);
assign packet_real_we=(state==STATE_HEADER)?(header_we):(copy_write_strobe);

// HEADER-RELEATED
	reg [1:0] headerSubstate;

assign header_done=(headerSubstate==2'h2 && header_addr==4'hF);
																				  
	wire [7:0] histogramSet, histogramSeq;
assign histogramSeq={5'h0,header_sel[2:0]};
assign histogramSet={5'h0,header_sel[5:3]};

always @(posedge clk) 
	if (state!=STATE_HEADER) headerSubstate<=0;
	else headerSubstate<=headerSubstate+1;

always @(posedge clk) header_we<=(state==STATE_HEADER) && (headerSubstate==2'h1 || headerSubstate==2'h2);

	wire [11:0] payloadSize;
assign payloadSize=packet_real_len-32;

always @(posedge clk) 
	if (state!=STATE_HEADER) header_addr<=0;
	else if (headerSubstate==2'h3) header_addr<=header_addr+1;

always @(posedge clk)  begin
   case (header_addr) 
	4'h0 : header_data<=DestMAC[47:32];
	4'h1 : header_data<=DestMAC[31:16];
	4'h2 : header_data<=DestMAC[15:0];
	4'h3 : header_data<=OwnMAC[47:32];
	4'h4 : header_data<=OwnMAC[31:16];
	4'h5 : header_data<=OwnMAC[15:0];
	4'h6 : header_data<=PacketType;		
/// FROM HERE DOWN, WE BYTE-SWAP for Linux
	4'h7 : header_data<={4'h2,4'h1,4'h0,4'h0}; // Retransmit,Subtype,Type,Version
	4'h8 : header_data<={hlxId,8'h0}; 
	4'h9 : header_data<={startOrbit[7:0],startOrbit[15:8]};
	4'ha : header_data<={startOrbit[23:16],startOrbit[31:24]};
	4'hb : header_data<={finishOrbit[7:0],finishOrbit[15:8]};
	4'hc : header_data<={finishOrbit[23:16],finishOrbit[31:24]};
	4'hd : header_data<={histogramSet,histogramSeq};
	4'he : header_data<={payloadSize[7:0],packet_fifo[3:0],payloadSize[11:8]};
	4'hf : header_data<={16'hFFFF};
	endcase
end


/// COPY-RELATED

	reg [2:0] copySubstate;

   wire [8:0] inc_histo_addr;

assign inc_histo_addr=histogram_addr[8:0]+1;

always @(posedge clk)
	if (state!=STATE_COPY) copySubstate<=3'h0; // override
	else begin
		case (copySubstate) 
		3'h0 : begin 
				   if (state==STATE_COPY) begin 
						copySubstate<=3'h1;
  						histogram_addr<={1'h0,header_sel,9'h0};
					end
					else begin
					  copySubstate<=3'h0;
					  histogram_addr<=15'h0;
					end
				 end
		3'h1 : if (histogram_read_ack) copySubstate<=3'h2;		 // raise the read-side strobe and wait for ack
		       else copySubstate<=3'h1;
		3'h2 : copySubstate<=3'h3; /// hold the write-side strobe						  
		3'h3 : copySubstate<=3'h4; /// hold the write-side strobe						  
		3'h4 : if (!histogram_read_ack) copySubstate<=3'h5; /// lower the write-side strobe and read-side strobe
		       else copySubstate<=3'h4;
	   3'h5 : begin // update address cnd check for end of copy	 
		         if (payloadSize[9:1]==inc_histo_addr) copySubstate<=3'h6; // end of copy
					else copySubstate<=3'h1; // continue copy
				 	histogram_addr[8:0]<=inc_histo_addr;
					histogram_addr[15:9]<=histogram_addr[15:9];
				 end
		3'h6 : if (state!=STATE_COPY) copySubstate<=3'h0;
				 else copySubstate<=3'h6; // wait for done
		default : copySubstate<=3'h0;
		endcase
	end

assign histogram_read_strobe = (copySubstate==3'h1 || copySubstate==3'h2 || copySubstate==3'h3);
assign copy_write_strobe = (copySubstate==3'h2 || copySubstate==3'h3);
assign copy_done = (copySubstate==3'h6);
assign copy_addr = histogram_addr[8:0];

/// XMIT-RELEATED

	reg [2:0] xmit_timer;
parameter XMIT_TIMER_DONE=3'h7;
parameter XMIT_TIMER_PULSE=3'h4;

assign packet_real_send=(state==STATE_XMIT)&&(xmit_timer<XMIT_TIMER_PULSE);

always @(posedge clk)
	if (state!=STATE_XMIT) xmit_timer<=0;
	else if (xmit_timer<XMIT_TIMER_DONE) xmit_timer<=xmit_timer+1;

/// MASTER STATE MACHINE

always @(posedge clk)
	if (reset) state<=STATE_IDLE; 
	else if (state==STATE_IDLE && packet_send) state<=STATE_COPY;
	else if (state==STATE_COPY && copy_done) state<=STATE_HEADER;
	else if (state==STATE_HEADER && header_done) state<=STATE_XMIT;
	/*
	else if (state==STATE_XMIT && xmit_timer==XMIT_TIMER_DONE && !packet_send_busy) begin
		if (mode_wait_send) state<=STATE_WAIT;
		else state<=STATE_IDLE;
   end
	else if (state==STATE_WAIT && packet_send_completed) state<=STATE_IDLE;
	*/
	else if (state==STATE_XMIT && xmit_timer==XMIT_TIMER_DONE) state<=STATE_IDLE;
  	else state<=state;

/*
always @(posedge clk)
	if (state!=STATE_WAIT) begin
		packet_send_completed<=0;
	   packet_send_poll<=0;
	end else begin
		if (!packet_send_poll) begin
			packet_send_poll<=1;
	   end else if (!packet_send_busy) begin
			packet_send_poll<=0;
			packet_send_completed<=!packet_send_macbusy;
		end
	end
 */

always @(posedge clk)
	if (reset) packet_fifo<=0;
	else if (state==STATE_XMIT && xmit_timer==XMIT_TIMER_DONE) begin
		if (packet_fifo==MAX_PACKET_COUNT) packet_fifo<=0;
		else packet_fifo<=packet_fifo+1;
	end else packet_fifo<=packet_fifo;

always @(posedge clk)
	if (state==STATE_IDLE && packet_send) begin
	   header_sel<=packet_sel;
		packet_real_len<=packet_len+32; // header added on
//		packet_fifo_save<=packet_fifo;
   end

endmodule
